import io
from collections import defaultdict, OrderedDict
from typing import Any


class Item(OrderedDict):
    def __init__(self, *args, **kwargs):
        self.text = kwargs.pop('text', '').strip()
        _dic = kwargs.pop('attributes', None)
        self.attributes = _dic if _dic is not None and isinstance(_dic, dict) else defaultdict(str)
        super(Item, self).__init__(*args, **kwargs)

    def filtered_attributes(self, ignore=()):
        return {k: v for k, v in self.attributes.items() if k not in ignore}

    def __repr__(self):
        return self.text

    def __getattr__(self, item: str) -> Any:
        if item in self.__dict__:
            return self.__dict__[item]
        return self.attributes.get(item.lower(), '')

    @staticmethod
    def odict2item(value):
        d = Item()

        if isinstance(value, OrderedDict):
            for k, v in value.items():
                if isinstance(v, OrderedDict):  # assumes v is also list of pairs
                    d[k] = [Item.odict2item(v)]
                elif isinstance(v, list):
                    d[k] = [Item.odict2item(_v) for _v in v]
                elif isinstance(v, str):
                    if k.startswith('@'):  # an attribute
                        d.attributes[k[1:].lower()] = v
                    elif k.startswith('#'):  # the text element
                        d.text = v
                    else:
                        d[k] = Item.odict2item(v)
                else:  # an empty tag
                    d[k] = ''
            return d
        elif isinstance(value, str):
            d.text = value
            return d
        return value


class XML(Item):
    @classmethod
    def parse_xml(cls, filename, namespaces={}):
        import xmltodict
        with io.open(filename, 'r', encoding='utf-8') as fp:
            x = xmltodict.parse(fp.read(), process_namespaces=True if namespaces else False, namespaces=namespaces)

            root_keys = list(x.keys())

            if not root_keys:  # no root
                return cls()

            root = Item.odict2item(x[root_keys[0]])
            return cls(**{**root, **root.__dict__})



import collections


def find_child(data, path, default_value=None, explicit_list=False):
	next_list = False
	for i, folder in enumerate(path):
		if next_list:
			next_list = False
			try:
				data = data[i]
			except:
				return default_value
		if type(data) is dict:
			#dictionary
			try:
				data = data[folder]
				if len(path) == i +1:
					return data
			except:
				return default_value
		elif type(data) is list and not explicit_list:
			#list
			r = []
			for x in data:
				res = find_child(x, path[i:])
				if res is not None:
					r.append(res)
			if len(r) == 0:
				return [default_value]
			else:
				return r
		elif type(data) is list and explicit_list and folder is list:
			next_list = True
		else:
			return default_value
	return default_value

def find_parent(data, path, default_value=None, explicit_list=False):
	if explicit_list:
		if path[-2] is list:
			return find_child(data, path[:-2])
		else:
			return find_child(data, path[:-1])
	else:
		return find_child(data, path[:-1])









from collections import Counter
from itertools import groupby
from tqdm import tqdm


def strtype(v): return str(type(v).__name__)


class Container:
    def __init__(self, data):
        self.data = data
        self.type = strtype(data)
        self.contains = []
        self.inner_content = []

        if isinstance(data, dict):
            values = data.items()
            self.contains = [(v[0], strtype(v[1]),) for v in values]
            self.inner_content = [Container(v[1]) for v in values if strtype(v[1]) in ['dict', 'list', 'tuple', 'set']]
        elif isinstance(data, list) or isinstance(data, tuple) or isinstance(data, set):
            self.contains = [strtype(v) for v in data]
            self.inner_content = [Container(v) for v in self.data if strtype(v) in ['dict', 'list', 'tuple', 'set']]
        else:  # do nothing
            pass

        self.contains_counter = Counter(self.contains)

    def __repr__(self, depth=0, _indent=4, content_only=False, show_progress=True):
        _output = ''
        if not content_only:
            _output += " " * _indent * depth + f"{self.type}:\n"

        _t = tqdm
        if show_progress is False:
            _t = list

        for k, v in _t(self.contains_counter.items()):
            if isinstance(k, tuple):  # a dictionary
                _k, _type = k
                _output += " " * _indent * (depth + 1) + f"{_k}:"
                if _type in ['dict', 'list', 'tuple', 'set']:
                    _output += "\n" + Container(self.data[_k]).__repr__(depth + 2, content_only=True,
                                                                        show_progress=False)
                else:
                    _output += f" {_type}\n"
            else:
                _output += " " * _indent * (depth + 1) + f"{k}: ({v})\n"
                _inner_containers = [_i for _i in self.inner_content if _i.type == k]
                if len(_inner_containers) > 0:
                    _inner_containers = groupby(sorted(_inner_containers, key=lambda __k: __k.type),
                                                key=lambda __k: __k.type)  # group them based on their type
                    for _k, groups in _inner_containers:
                        # group them based on their content type
                        _inner_inner_containers = groupby(sorted(groups, key=lambda __k: __k.contains),
                                                          key=lambda __k: __k.contains)
                        for __k, __groups in _inner_inner_containers:
                            _data = list(__groups)
                            _output += " " * _indent * (depth + 2) + f"{len(_data)}:\n"
                            _output += _data[0].__repr__(depth=depth + 2, content_only=True, show_progress=False)

        return _output


def summarize(data):
    return Container(data).__repr__(0)
